import os
from dataclasses import asdict, dataclass

import numpy as np
import pandas as pd
from omegaconf import OmegaConf
from torch.utils.tensorboard import SummaryWriter


__all__ = [
    "TrainerState",
    "AverageMeter",
    "MetricsTracker",
    "CSVLogger",
    "write_to_tensorboard",
    "resolve_logdir",
]


@dataclass
class TrainerState:
    """
    A Trainer state is a container for holding specific updating values in the training process and is saved when
    checkpointing.

    Args:
        epoch: Current epoch number
        total_epochs: Total epochs to train the model
        global_step: Number of the update steps so far, one step is a full training step (one batch)
        metric_for_best_checkpoint: The metric key for choosing the best checkpoint (Also given in the TrainerConfig)
        best_metric_value: The value of the best checkpoint saved so far
        best_checkpoint: Path to the best model checkpoint so far
        logs_dir: Path to the logs directory
    """
    epoch: int = 1
    total_epochs: int = None
    global_step: int = 0
    metric_for_best_checkpoint: str = None
    best_metric_value: float = None
    best_checkpoint: str = None
    logs_dir: str = None

    def update(self, items: dict, **kwargs):
        items.update(kwargs)
        for k, v in items.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def update_best_results(self, metric_value, objective, step):
        if objective == "maximize":
            operator = np.greater
        elif objective == "minimize":
            operator = np.less
        else:
            raise ValueError(f"`objective` must be either `maximize` or `minimize`, got `{objective}`!")

        if self.best_metric_value is None:
            self.best_metric_value = metric_value
            self.best_checkpoint = step

        elif operator(metric_value, self.best_metric_value):
            self.best_metric_value = metric_value
            self.best_checkpoint = step

    def save(self, path, drop_none: bool = False):
        """
        Save the state to a .yaml file at `path`
        """
        state = asdict(self)
        if drop_none:
            state = {k: v for k, v in state.items() if v is not None}
        os.makedirs(os.path.dirname(path), exist_ok=True)
        OmegaConf.save(state, path)

    @classmethod
    def load(cls, path):
        """
        Load a trainer state from `path`
        """
        state_file = OmegaConf.load(path)
        state_dict = OmegaConf.to_container(state_file)
        state = cls(**state_dict)
        return state


class AverageMeter:
    """Compute and store the average and current value"""

    def __init__(self, name, fmt=":f"):
        self.name = name
        self.fmt = fmt
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = "{name} {val" + self.fmt + "} ({avg" + self.fmt + "})"
        return fmtstr.format(**self.__dict__)


class MetricsTracker:
    def __init__(self, metrics):
        self.metrics = metrics or []
        self.trackers = {}
        if len(self.metrics):
            for m in self.metrics.values():
                for metric_key in m.config.output_keys:
                    self.trackers[metric_key] = AverageMeter(metric_key)
        if "loss" not in self.trackers:
            self.trackers["loss"] = AverageMeter("loss")

    def update(self, results):
        for metric_name, tracker in self.trackers.items():
            tracker.update(results[metric_name])

    def reset(self):
        for tracker in self.trackers.values():
            tracker.reset()

    def avg(self):
        avg_results = {}
        for metric_name, tracker in self.trackers.items():
            avg_results[metric_name] = tracker.avg

        return avg_results


class CSVLogger:
    def __init__(self, logs_dir: str, csv_filename: str):
        self.save_path = os.path.join(logs_dir, csv_filename)
        self.df = pd.DataFrame({})

    def write(self, logs: dict, step: int):
        all_logs = {"step": step}
        all_logs.update({k: [v] for k, v in logs.items()})
        row = pd.DataFrame(all_logs)
        self.df = pd.concat([self.df, row])
        self.df.to_csv(self.save_path, index=False)


def write_to_tensorboard(writer: SummaryWriter, logs: dict, step: int):
    for metric_name, value in logs.items():
        writer.add_scalar(metric_name, value, step)


def resolve_logdir(log_dir) -> str:
    import socket
    from datetime import datetime

    current_time = datetime.now().strftime("%b%d_%H-%M-%S")
    return os.path.join(log_dir, current_time + "_" + socket.gethostname())



from __future__ import annotations

from typing import Dict, List

import numpy as np
import torch

from ....constants import Backends
from ....registry import register_model
from ....utils import is_backend_available
from ...model import Model
from .vit_config import ViTConfig


if is_backend_available(Backends.TRANSFORMERS):
    from transformers import ViTConfig as ViTConfig_
    from transformers import ViTModel

if is_backend_available(Backends.PILLOW):
    from PIL import Image

_required_backends = [Backends.TRANSFORMERS, Backends.TOKENIZERS, Backends.PILLOW]


@register_model("vit", config_class=ViTConfig)
class ViT(Model):
    required_backends = _required_backends
    image_processor = "image_processor"
    loss_func_name = "cross_entropy"

    def __init__(self, config: ViTConfig, **kwargs):
        super().__init__(config=config, **kwargs)
        self.vit = ViTModel(ViTConfig_(**self.config))

    def forward(
        self,
        pixel_values=None,
        bool_masked_pos=None,
        head_mask=None,
        output_attentions=None,
        output_hidden_states=None,
        interpolate_pos_encoding=None,
    ):
        outputs = self.vit(
            pixel_values=pixel_values,
            bool_masked_pos=bool_masked_pos,
            head_mask=head_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            interpolate_pos_encoding=interpolate_pos_encoding,
        )

        return outputs

    def preprocess(self, inputs: List[str] | List[np.ndarray] | List["Image"] | List[torch.Tensor], **kwargs):
        image_processor = self.preprocessor[self.image_processor]
        processed_outputs = image_processor(inputs, **kwargs)
        return processed_outputs

    def post_process(self, model_outputs: Dict[str, torch.Tensor]):
        outputs = {
            "last_hidden_state": model_outputs.get("last_hidden_state", None),
            "pooler_output": model_outputs.get("pooler_output", None),
            "hidden_states": model_outputs.get("hidden_states", None),
            "attentions": model_outputs.get("attentions", None),
        }
        return outputs



"""
Configs are at the core of Hezar. All core modules like `Model`, `Preprocessor`, `Trainer`, etc. take their parameters
as a config container which is an instance of `Config` or its derivatives. A `Config` is a Python dataclass with
auxiliary methods for loading, saving, uploading to the hub, etc.

Examples:
    >>> from hezar.configs import ModelConfig
    >>> config = ModelConfig.load("hezarai/bert-base-fa")

    >>> from hezar.models import BertMaskFillingConfig
    >>> bert_config = BertMaskFillingConfig(vocab_size=50000, hidden_size=768)
    >>> bert_config.save("saved/bert", filename="model_config.yaml")
    >>> bert_config.push_to_hub("hezarai/bert-custom", filename="model_config.yaml")
"""
from __future__ import annotations

import os
import tempfile
from dataclasses import asdict, dataclass, field, fields
from enum import Enum
from pprint import pformat
from typing import Any, Dict, List, Literal, Optional, Tuple

from huggingface_hub import create_repo, hf_hub_download, upload_file
from omegaconf import DictConfig, OmegaConf

from .constants import (
    DEFAULT_MODEL_CONFIG_FILE,
    HEZAR_CACHE_DIR,
    ConfigType,
    LRSchedulerType,
    OptimizerType,
    PrecisionType,
    TaskType,
)
from .utils import Logger, get_module_config_class


__all__ = [
    "Config",
    "ModelConfig",
    "PreprocessorConfig",
    "TrainerConfig",
    "DatasetConfig",
    "EmbeddingConfig",
    "MetricConfig",
]

logger = Logger(__name__)

CONFIG_CLASS_VARS = ["name", "config_type"]

_config_to_type_mapping = {
    "ModelConfig": ConfigType.MODEL,
    "PreprocessorConfig": ConfigType.PREPROCESSOR,
    "TrainerConfig": ConfigType.TRAINER,
    "DatasetConfig": ConfigType.DATASET,
    "EmbeddingConfig": ConfigType.EMBEDDING,
    "CriterionConfig": ConfigType.CRITERION,
    "MetricConfig": ConfigType.METRIC,
}
_type_to_config_mapping = {v: k for k, v in _config_to_type_mapping.items()}


@dataclass
class Config:
    """
    Base class for all configs in Hezar.

    All configs are simple dataclasses that have some customized functionalities to manage their attributes. There are
    also some Hezar specific methods: `load`, `save` and `push_to_hub`.

    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.BASE)

    def __post_init__(self):
        # Class variables cannot be init-able
        fields_dict = {f.name: f for f in fields(self)}
        for attr in CONFIG_CLASS_VARS:
            if fields_dict[attr].init == True:  # noqa
                raise ValueError(
                    f"The parameter `{attr}` in a config should be either non-initable or unannotated! "
                    f"\nYou should define it as either:\n"
                    f"`{attr} = '{getattr(self, attr)}'`"
                    f" or "
                    f"`{attr}: str = field(default='{getattr(self, attr)}', init=False)`"
                )

        # Convert enums to values
        for param, value in self.dict().items():
            if isinstance(getattr(self, param), Enum):
                setattr(self, param, str(getattr(self, param)))

    def __str__(self):
        return pformat(self.dict())

    def __getitem__(self, item):
        try:
            return self.dict()[item]
        except KeyError:
            raise AttributeError(f"`{self.__class__.__name__}` does not have the parameter `{item}`!")

    def __len__(self):
        return len(self.dict())

    def __iter__(self):
        return iter(self.dict())

    @classmethod
    def fields(cls):
        return cls.__dataclass_fields__

    def dict(self):
        """
        Returns the config object as a dictionary (works on nested dataclasses too)

        Returns:
            The config object as a dictionary
        """
        return asdict(self)

    def keys(self):
        return list(self.dict().keys())

    def get(self, key, default=None):
        return getattr(self, key, default)

    def update(self, d: dict, **kwargs):
        """
        Update config with a given dictionary or keyword arguments. If a key does not exist in the attributes, prints a
        warning but sets it anyway.

        Args:
            d: A dictionary
            **kwargs: Key/value pairs in the form of keyword arguments

        Returns:
            The config object itself but the operation happens in-place anyway
        """
        d.update(kwargs)
        for k, v in d.items():
            if k not in self.fields():
                logger.warning(f"`{str(self.__class__.__name__)}` does not take `{k}` as a config parameter!")
            setattr(self, k, v)
        return self

    @classmethod
    def load(
        cls,
        hub_or_local_path: str | os.PathLike,
        filename: Optional[str] = None,
        subfolder: Optional[str] = None,
        repo_type: str = None,
        cache_dir: str = None,
        **kwargs,
    ) -> "Config":
        """
        Load config from Hub or locally if it already exists on disk (handled by HfApi)

        Args:
            hub_or_local_path: Local or Hub path for the config
            filename: Configuration filename
            subfolder: Optional subfolder path where the config is in
            repo_type: Repo type e.g, model, dataset, etc
            cache_dir: Path to cache directory
            **kwargs: Manual config parameters to override

        Returns:
            A Config instance
        """
        filename = filename or DEFAULT_MODEL_CONFIG_FILE
        subfolder = subfolder or ""

        config_path = os.path.join(hub_or_local_path, subfolder, filename)
        is_local = os.path.isfile(config_path)
        if os.path.isdir(hub_or_local_path) and not is_local:
            raise EnvironmentError(
                f"Path `{hub_or_local_path}` exists locally but the config file {filename} is missing!"
            )
        # if the file or repo_id does not exist locally, load from the Hub
        if not is_local:
            config_path = hf_hub_download(
                hub_or_local_path,
                filename=filename,
                subfolder=subfolder,
                cache_dir=cache_dir or HEZAR_CACHE_DIR,
                repo_type=repo_type,
            )
        # Load config file and convert to dictionary
        dict_config = OmegaConf.load(config_path)
        config = OmegaConf.to_container(dict_config)
        # Check if config_type in the file and class are equal
        config_type = config.get("config_type", ConfigType.BASE)
        if config_type in _config_to_type_mapping.values():
            if config_type != cls.config_type:
                raise ValueError(
                    f"The `config_type` for `{cls.__name__}` is `{cls.config_type}` "
                    f"which is different from the `config_type` parameter in `{filename}` which is `{config_type}`!"
                )
        config_cls = get_module_config_class(config["name"], registry_type=config_type)
        if config_cls is None:
            config_cls = cls
        config = config_cls.from_dict(config, **kwargs)
        return config

    @classmethod
    def from_dict(cls, dict_config: Dict | DictConfig, **kwargs):
        """
        Load config from a dict-like object. Nested configs are also recursively converted to their classes if possible.
        """
        # Update config parameters with kwargs
        dict_config.update(**kwargs)

        for k, v in dict_config.items():
            if isinstance(v, Dict) and "name" in v and "config_type" in v:
                config_cls = get_module_config_class(v["name"], v["config_type"])
                if config_cls is not None:
                    dict_config[k] = config_cls.from_dict(v)

        dict_config = {k: v for k, v in dict_config.items() if k in cls.fields() and cls.fields()[k].init}

        config = cls(**dict_config)  # noqa

        return config

    def save(
        self,
        save_dir: str | os.PathLike,
        filename: str,
        subfolder: Optional[str] = None,
        skip_none_fields: Optional[bool] = True,
    ):
        """
        Save the `*config.yaml` file to a local path

        Args:
             save_dir: Save directory path
             filename: Config file name
             subfolder: Subfolder to save the config file
             skip_none_fields (bool): Whether to skip saving None values or not
        """
        subfolder = subfolder or ""
        config = self.dict()

        if skip_none_fields:
            # exclude None items
            config = {k: v for k, v in config.items() if v is not None}

        # make and save to directory
        os.makedirs(os.path.join(save_dir, subfolder), exist_ok=True)
        save_path = os.path.join(save_dir, subfolder, filename)
        OmegaConf.save(config, save_path)

        return save_path

    def push_to_hub(
        self,
        repo_id: str,
        filename: str,
        subfolder: Optional[str] = None,
        repo_type: Optional[str] = "model",
        skip_none_fields: Optional[bool] = True,
        private: Optional[bool] = False,
        commit_message: Optional[str] = None,
    ):
        """
        Push the config file to the hub

        Args:
            repo_id (str): Repo name or id on the Hub
            filename (str): config file name
            subfolder (str): subfolder to save the config
            repo_type (str): Type of the repo e.g, model, dataset, space
            skip_none_fields (bool): Whether to skip saving None values or not
            private (bool): Whether the repo type should be private or not (ignored if the repo exists)
            commit_message (str): Push commit message
        """
        path_in_repo = f"{subfolder}/{filename}" if subfolder else filename
        subfolder = subfolder or ""

        # create remote repo
        create_repo(repo_id, repo_type=repo_type, private=private, exist_ok=True)
        # save to tmp and prepare for push
        cache_path = tempfile.mkdtemp()
        config_path = self.save(cache_path, filename=filename, subfolder=subfolder, skip_none_fields=skip_none_fields)
        # push to hub
        if commit_message is None:
            commit_message = f"Hezar: Upload {filename}"
        upload_file(
            path_or_fileobj=config_path,
            path_in_repo=path_in_repo,
            repo_id=repo_id,
            repo_type=repo_type,
            commit_message=commit_message,
        )
        logger.log_upload_success(name=f"{self.__class__.__name__}()", target_path=os.path.join(repo_id, path_in_repo))


@dataclass
class ModelConfig(Config):
    """
    Base dataclass for all model configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.MODEL)


@dataclass
class PreprocessorConfig(Config):
    """
    Base dataclass for all preprocessor configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.PREPROCESSOR)


@dataclass
class DatasetConfig(Config):
    """
    Base dataclass for all dataset configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.DATASET)
    task: TaskType | List[TaskType] = field(
        default=None, metadata={"help": "Name of the task(s) this dataset is built for"}
    )
    path: str = None


@dataclass
class EmbeddingConfig(Config):
    """
    Base dataclass for all embedding configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.EMBEDDING)
    bypass_version_check: bool = field(
        default=False,
        metadata={"help": "Whether to bypass checking gensim/numpy/hezar version compatibility"},
    )


@dataclass
class MetricConfig(Config):
    """
    Base dataclass config for all metric configs
    """

    name: str = field(init=False, default=None)
    config_type: str = field(init=False, default=ConfigType.METRIC)
    objective: Literal["maximize", "minimize"] = None
    output_keys: List | Tuple = None
    n_decimals: int = 4


@dataclass
class TrainerConfig(Config):
    """
    Base dataclass for all trainer configs

    Args:
        task (str, TaskType):
            The training task. Must be a valid name from `TaskType`.
        output_dir (str):
            Path to the directory to save trainer properties.
        device (str):
            Hardware device e.g, `cuda:0`, `cpu`, etc.
        num_epochs (int):
            Number of total epochs to train the model.
        init_weights_from (str):
            Path to a model from disk or Hub to load the initial weights from.
        num_dataloader_workers (int):
            Number of dataloader workers, defaults to 4 .
        seed (int):
            Control determinism of the run by setting a seed value. defaults to 42.
        optimizer (OptimizerType):
            Name of the optimizer, available values include properties in `OptimizerType` enum.
        learning_rate (float):
            Initial learning rate for the optimizer.
        weight_decay (float):
            Optimizer weight decay value.
        lr_scheduler (LRSchedulerType):
            Optional learning rate scheduler among `LRSchedulerType` enum.
        lr_scheduler_kwargs (Dict[str, Any]):
            LR scheduler instructor kwargs depending on the scheduler type
        batch_size (int):
            Training batch size.
        eval_batch_size (int):
            Evaluation batch size, defaults to `batch_size` if None.
        gradient_accumulation_steps (int):
            Number of updates steps to accumulate before performing a backward/update pass, defaults to 1.
        distributed (bool):
            Whether to use distributed training or not (via the `accelerate` package)
        mixed_precision (PrecisionType | str):
            Mixed precision type e.g, fp16, bf16, etc. (disabled by default)
        evaluate_with_generate (bool):
            Whether to use `generate()` in the evaluation step or not. (only applicable for generative models).
        metrics (List[str | MetricConfig]):
            A list of metrics. Depending on the `valid_metrics` in the specific MetricsHandler of the Trainer.
        metric_for_best_model (str):
            Reference metric key to watch for determining the best model. Recommended to have a {train. | evaluation.}
            prefix (e.g, evaluation.f1, train.accuracy, etc.) but if not, defaults to
            `evaluation.{metric_for_best_model}`.
        save_freq (int):
            Save the trainer stats and everything every `save_freq` epochs.
        checkpoints_dir (str):
            Path to the checkpoints' folder. The actual files will be saved under `{output_dir}/{checkpoints_dir}`.
        logs_dir (str):
            Path to the logs' folder. The actual log files will be saved under `{output_dir}/{logs_dir}`.
    """

    name: str = field(init=False, default="trainer")
    config_type: str = field(init=False, default=ConfigType.TRAINER)
    output_dir: str
    task: str | TaskType
    device: str = "cuda"
    num_epochs: int = None
    init_weights_from: str = None
    num_dataloader_workers: int = 0
    dataloader_drop_last: bool = False
    dataloader_shuffle: bool = True
    seed: int = 42
    optimizer: str | OptimizerType = None
    learning_rate: float = 2e-5
    weight_decay: float = 0.0
    lr_scheduler: str | LRSchedulerType = None
    lr_scheduler_kwargs: Dict[str, Any] = None
    batch_size: int = None
    eval_batch_size: int = None
    gradient_accumulation_steps: int = 1
    distributed: bool = field(
        init=False,
        default=False,
        metadata={"help": "Distributed training isn't supported yet! We're working hard to solve some bugs rn :("}
    )
    mixed_precision: PrecisionType | str | None = None
    use_cpu: bool = False
    evaluate_with_generate: bool = True
    metrics: List[str | MetricConfig] = None
    metric_for_best_model: str = "evaluation.loss"
    save_freq: int = 1
    checkpoints_dir: str = "checkpoints"
    logs_dir: str = "logs"

    def __post_init__(self):
        """
        Perform some argument sanitization and filtering here to avoid unexpected behavior in the trainer.
        The need for having this method is that some fields in the Trainer's config have correlations with each other
        and not controlling them can lead to conflicts.
        """
        super().__post_init__()
        if self.task not in list(TaskType):
            raise ValueError(
                f"Invalid task `{self.task}` passed to `TrainerConfig`. "
                f"Available options are {TaskType.list()}",
            )
        if not (self.metric_for_best_model.startswith("evaluation") or self.metric_for_best_model.startswith("train")):
            self.metric_for_best_model = f"evaluation.{self.metric_for_best_model}"



from dataclasses import dataclass

import torch
from datasets import load_dataset

from ...builders import build_preprocessor
from ...configs import DatasetConfig
from ...constants import Backends, TaskType
from ...preprocessors import ImageProcessorConfig, Tokenizer
from ...registry import register_dataset
from ...utils import Logger, shift_tokens_right
from ..data_collators import ImageCaptioningDataCollator
from .dataset import Dataset


logger = Logger(__name__)

_required_backends = [Backends.SCIKIT]


@dataclass
class ImageCaptioningDatasetConfig(DatasetConfig):
    """
    Configuration class for image captioning datasets.

    Args:
        path (str): Path to the dataset.
        tokenizer_path (str): Path to the tokenizer file.
        text_column (str): Column name for text in the dataset.
        images_paths_column (str): Column name for image paths in the dataset.
        max_length (int): Maximum length of text.
        test_split_size (float): Size of the test split.
        image_processor_config (ImageProcessorConfig): Configuration for image processing.

    """
    name = "image_captioning"
    task: TaskType = TaskType.IMAGE2TEXT
    path: str = None
    tokenizer_path: str = None
    text_column: str = "label"
    images_paths_column = "image_path"
    max_length: int = None
    test_split_size: float = 0.2
    image_processor_config: ImageProcessorConfig = None


@register_dataset("image_captioning", config_class=ImageCaptioningDatasetConfig)
class ImageCaptioningDataset(Dataset):
    required_backends = _required_backends

    def __init__(self, config: ImageCaptioningDatasetConfig, split=None, **kwargs):
        super().__init__(config=config, split=split, **kwargs)
        self.data = self._load(split)
        self.image_processor = build_preprocessor("image_processor", config=self.config.image_processor_config)
        self.tokenizer = Tokenizer.load(self.config.tokenizer_path)
        self.data_collator = ImageCaptioningDataCollator(
            self.tokenizer,
            padding_type="max_length" if self.config.max_length is not None else "longest",
            max_length=self.config.max_length
        )

    def __len__(self):
        """
        Returns the length of the dataset.

        Returns:
            int: The length of the dataset.

        """
        return len(self.data)

    def _load(self, split=None):
        """
        Load the dataset and clean up invalid samples.

        Args:
            split: Dataset split, defaults to None.

        Returns:
            Dataset: The cleaned dataset.

        """
        data = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir)
        return data

    def __getitem__(self, index):
        """
        Get a specific item from the dataset.

        Args:
            index: Index of the item to retrieve.

        Returns:
            dict: The input data.
        """
        path, text = self.data[index].values()
        pixel_values = self.image_processor(path, return_tensors="pt")["pixel_values"]
        tokenized_inputs = self.tokenizer(text, padding="max_length", max_length=self.config.max_length)
        labels = torch.tensor([tokenized_inputs["token_ids"]])
        attention_mask = torch.tensor([tokenized_inputs["attention_mask"]])
        decoder_input_ids = shift_tokens_right(
            labels,
            pad_token_id=self.tokenizer.pad_token_id,
            decoder_start_token_id=self.tokenizer.bos_token_id,
        )
        inputs = {
            "pixel_values": pixel_values,
            "labels": labels,
            "decoder_input_ids": decoder_input_ids,
            "decoder_attention_mask": attention_mask,
        }
        return inputs



from __future__ import annotations

import os
import random
from typing import TYPE_CHECKING, Any, Callable, Dict, Tuple

import numpy as np
import pandas as pd
import torch
from huggingface_hub import create_repo, hf_hub_download, upload_file
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from .metrics_handlers import (
    Image2TextMetricHandler,
    MetricsHandler,
    SequenceLabelingMetricsHandler,
    SpeechRecognitionMetricsHandler,
    TextClassificationMetricsHandler,
    TextGenerationMetricsHandler,
)
from .trainer_utils import CSVLogger, TrainerState, resolve_logdir, write_to_tensorboard
from ..configs import TrainerConfig
from ..constants import (
    DEFAULT_DATASET_CONFIG_FILE,
    DEFAULT_TRAINER_CONFIG_FILE,
    DEFAULT_TRAINER_CSV_LOG_FILE,
    DEFAULT_TRAINER_STATE_FILE,
    DEFAULT_TRAINER_SUBFOLDER,
    HEZAR_CACHE_DIR,
    TQDM_BAR_FORMAT,
    Backends,
    LRSchedulerType,
    OptimizerType,
    TaskType,
)
from ..data.datasets import Dataset
from ..models import Model
from ..preprocessors import Preprocessor, PreprocessorsContainer
from ..utils import Logger, colorize_text, is_backend_available, sanitize_function_parameters, verify_dependencies

if TYPE_CHECKING:
    from accelerate import Accelerator

logger = Logger(__name__)

optimizers = {
    OptimizerType.ADAM: torch.optim.Adam,
    OptimizerType.ADAMW: torch.optim.AdamW,
    OptimizerType.SDG: torch.optim.SGD,
}
lr_schedulers = {
    LRSchedulerType.LAMBDA: torch.optim.lr_scheduler.LambdaLR,
    LRSchedulerType.STEP: torch.optim.lr_scheduler.StepLR,
    LRSchedulerType.MULTI_STEP: torch.optim.lr_scheduler.MultiStepLR,
    LRSchedulerType.ONE_CYCLE: torch.optim.lr_scheduler.OneCycleLR,
    LRSchedulerType.LINEAR: torch.optim.lr_scheduler.LinearLR,
    LRSchedulerType.EXPONENTIAL: torch.optim.lr_scheduler.ExponentialLR,
    LRSchedulerType.CYCLIC: torch.optim.lr_scheduler.CyclicLR,
    LRSchedulerType.SEQUENTIAL: torch.optim.lr_scheduler.SequentialLR,
    LRSchedulerType.POLYNOMIAL: torch.optim.lr_scheduler.PolynomialLR,
    LRSchedulerType.COSINE_ANEALING: torch.optim.lr_scheduler.CosineAnnealingLR,
}

task_to_metrics_handlers_mapping = {
    TaskType.TEXT_CLASSIFICATION: TextClassificationMetricsHandler,
    TaskType.SEQUENCE_LABELING: SequenceLabelingMetricsHandler,
    TaskType.IMAGE2TEXT: Image2TextMetricHandler,
    TaskType.SPEECH_RECOGNITION: SpeechRecognitionMetricsHandler,
    TaskType.TEXT_GENERATION: TextGenerationMetricsHandler,
}


class Trainer:
    """
    Base trainer class for training all Hezar models and all tasks. Usually you can use this class as-is, but for special
    cases you can also override any of the core methods in your own custom Trainer.

    Args:
        model ([`Model`] or `torch.nn.Module`): The main model to train and evaluate
        config (TrainerConfig): Training configuration and parameters
        train_dataset (Dataset): Train dataset
        eval_dataset (Dataset): Evaluation dataset
        data_collator: Collate function, usually included in the dataset object itself
        preprocessor: Preprocessor object(s)
        metrics_handler: Optional metrics handler
        optimizer (optim.Optimizer): Model optimizer
        lr_scheduler: Optional learning-rate scheduler
        accelerator (Accelerator) : Accelerator object for a customized distributed environment
    """

    trainer_subfolder = DEFAULT_TRAINER_SUBFOLDER
    trainer_config_file = DEFAULT_TRAINER_CONFIG_FILE
    trainer_csv_log_file = DEFAULT_TRAINER_CSV_LOG_FILE
    dataset_config_file = DEFAULT_DATASET_CONFIG_FILE
    trainer_state_file = DEFAULT_TRAINER_STATE_FILE
    default_optimizer = OptimizerType.ADAM
    default_lr_scheduler = None
    _required_backends = []

    def __init__(
        self,
        model: Model = None,
        config: TrainerConfig = None,
        train_dataset: Dataset = None,
        eval_dataset: Dataset = None,
        data_collator: Callable = None,
        preprocessor: Preprocessor | PreprocessorsContainer = None,
        metrics_handler: MetricsHandler = None,
        optimizer: torch.optim.Optimizer = None,
        lr_scheduler=None,
        accelerator: "Accelerator" = None,
    ):
        # Check if all required dependencies are installed
        verify_dependencies(self._required_backends)

        self.config = config
        self.device = "cuda" if torch.cuda.is_available() and not self.config.use_cpu else "cpu"

        # Set determinism
        self._set_seed(self.config.seed)

        # Setup model and preprocessor(s)
        self.model = self._setup_model(model)
        if self.model.preprocessor is None:
            if preprocessor is not None:
                self.model.preprocessor = preprocessor
            else:
                raise ValueError(
                    "You must set a preprocessor for the model or pass the preprocessor parameter to the Trainer!"
                )

        # Configure datasets and data loaders
        self.train_dataset = train_dataset
        self.eval_dataset = eval_dataset
        self.data_collator = data_collator or self.train_dataset.data_collator
        self.train_dataloader, self.eval_dataloader = self._setup_dataloaders()

        # Setup optimizer and (optionally) lr scheduler
        self.optimizer, self.lr_scheduler = self._setup_optimizers(optimizer, lr_scheduler)

        # Setup accelerated objects if possible
        if accelerator is None and self.config.distributed and not self.config.use_cpu:
            self.accelerator = self._setup_accelerator()
            self.scaler = self.accelerator.scaler
            self.device = self.accelerator.device
        else:
            self.accelerator = accelerator
            enabled = True if (
                self.config.mixed_precision is not None and not (self.config.use_cpu or self.device == "cpu")
            ) else False
            self.scaler = torch.cuda.amp.GradScaler(enabled=enabled)

        # Setup metrics handler and inner trackers for the trainer
        self.metrics_handler = metrics_handler or self._setup_metrics_handler()

        # Configure checkpoints and logging directories
        self.logs_dir = resolve_logdir(os.path.join(self.config.output_dir, self.config.logs_dir))
        self.checkpoints_dir = os.path.join(self.config.output_dir, self.config.checkpoints_dir)

        # Setup logging properties
        self.tensorboard = SummaryWriter(log_dir=self.logs_dir)
        self.csv_logger = CSVLogger(logs_dir=self.logs_dir, csv_filename=self.trainer_csv_log_file)

        self.current_epoch = 1

        # Configure trainer state
        self.state = TrainerState(
            epoch=self.current_epoch,
            total_epochs=self.config.num_epochs,
            metric_for_best_checkpoint=self.config.metric_for_best_model,
            logs_dir=self.tensorboard.log_dir,
        )

    @staticmethod
    def _set_seed(seed):
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)

    def _setup_model(self, model: Model) -> Model:
        """
        Download the model from HuggingFace Hub if `init_weights_from` is given in the config. Load the model to the
        device and return it.
        """
        if model is None:
            raise ValueError("`model` must be given to the Trainer!")
        hub_path = self.config.init_weights_from
        if hub_path is not None:
            if os.path.isdir(hub_path):
                model_path = os.path.join(hub_path, model.model_filename)
            else:
                model_path = hf_hub_download(
                    hub_path,
                    filename=model.model_filename,
                    cache_dir=HEZAR_CACHE_DIR,
                    resume_download=True,
                )
            model.load_state_dict(torch.load(model_path, map_location="cpu"))
        model.to(self.device)
        return model

    def _setup_dataloaders(self) -> Tuple[DataLoader, DataLoader | None]:
        """
        Set up data loaders (train/eval) and return them.

        Returns:
             A tuple of train and eval dataloaders
        """
        if self.train_dataset is not None:
            train_dataloader = DataLoader(
                dataset=self.train_dataset,
                batch_size=self.config.batch_size,
                collate_fn=self.data_collator,
                num_workers=self.config.num_dataloader_workers,
                drop_last=self.config.dataloader_drop_last,
                shuffle=self.config.dataloader_shuffle,
            )
        else:
            raise ValueError("Cannot create train dataloader because `train_dataset` is not given!")
        if self.eval_dataset is not None:
            eval_dataloader = DataLoader(
                dataset=self.eval_dataset,
                batch_size=self.config.eval_batch_size or self.config.batch_size,
                collate_fn=self.data_collator,
                num_workers=self.config.num_dataloader_workers,
                drop_last=self.config.dataloader_drop_last,
                shuffle=self.config.dataloader_shuffle,
            )
        else:
            logger.warning(
                "Cannot create eval dataloader because `eval_dataset` is not given to the Trainer! "
                "Setting eval_dataloader to None..."
            )
            eval_dataloader = None

        return train_dataloader, eval_dataloader

    def _setup_optimizers(self, optimizer: torch.optim.Optimizer = None, lr_scheduler=None):
        """
        Set up the optimizer and lr lr_scheduler if they're not already given

        Args:
            optimizer: If None do nothing and return it, otherwise build it using the train config
            lr_scheduler: If None do nothing and return it, otherwise build it using the train config

        Returns:
            Optimizer and lr_scheduler
        """
        if optimizer is None:
            optimizer_type = self.config.optimizer or self.default_optimizer
            optimizer = optimizers[optimizer_type](
                self.model.parameters(),
                lr=self.config.learning_rate,
                weight_decay=self.config.weight_decay,
            )

            if lr_scheduler is None:
                scheduler_name = self.config.lr_scheduler or self.default_lr_scheduler
                scheduler_kwargs = self.config.lr_scheduler_kwargs or {}
                if scheduler_name is None:
                    lr_scheduler = None
                else:
                    lr_scheduler = lr_schedulers[scheduler_name](optimizer, **scheduler_kwargs, verbose=True)
        return optimizer, lr_scheduler

    def _setup_accelerator(self):
        if is_backend_available(Backends.ACCELERATE):
            from accelerate import Accelerator

            accelerator = Accelerator(
                mixed_precision=self.config.mixed_precision,
                step_scheduler_with_optimizer=True if self.lr_scheduler is not None else False,
            )
            self.model, self.optimizer, self.lr_scheduler, self.train_dataloader, self.eval_dataloader = (
                accelerator.prepare(
                    self.model,
                    self.optimizer,
                    self.lr_scheduler,
                    self.train_dataloader,
                    self.eval_dataloader,
                )
            )
        else:
            raise ValueError(
                "The configuration for this trainer requires the package `accelerate` to be installed! "
                "(config.distributed=True)"
            )

        return accelerator

    def _setup_metrics_handler(self):
        """
        Setup MetricsHandler instance for the trainer

        Returns:
            A MetricsHandler subclass instance based on self.config.task
        """
        metrics_handler_cls = task_to_metrics_handlers_mapping[self.config.task]
        metrics_handler = metrics_handler_cls(metrics=self.config.metrics, trainer=self)  # noqa
        return metrics_handler

    def load_from_checkpoint(self, checkpoint: str | bool = True, load_best: bool = False):
        """
        Load trainer states like model weights, optimizer, etc. from a checkpoint

        Args:
            checkpoint: Path to checkpoint directory
            load_best: Whether to load the best checkpoint or not, if False, loads the latest checkpoint
        """
        if os.path.isdir(checkpoint) and load_best:
            logger.warning("The `load_best` parameter has no effect when `checkpoint` is a path!")

        self.state = TrainerState.load(os.path.join(self.checkpoints_dir, self.trainer_state_file))
        if isinstance(checkpoint, bool):
            if load_best:
                checkpoint = os.path.join(self.checkpoints_dir, str(self.state.best_checkpoint))
            else:
                checkpoint = os.path.join(self.checkpoints_dir, str(self.state.epoch))
        if os.path.isdir(checkpoint):
            # Figure out the epoch number
            epoch = os.path.basename(checkpoint) if os.path.basename(checkpoint).isdigit() else self.state.epoch
            if str(epoch).isdigit():
                self.state.epoch = int(epoch)
            # Load model's state dict
            model_path = os.path.join(checkpoint, self.model.model_filename)
            if os.path.isfile(model_path):
                self.model.load_state_dict(torch.load(model_path))
                if self.accelerator is not None:
                    self.model = self.accelerator.prepare(self.model)
                logger.info(f"Successfully loaded checkpoint from `{checkpoint}` ")
            else:
                raise FileNotFoundError(
                    f"Could not find `{self.model.model_filename}` at `{os.path.dirname(model_path)}`!\n"
                )
        else:
            logger.warning(
                f"{checkpoint} does not seem to be a valid checkpoint!"
            )

    def load_csv_logs(self, logs_dir=None):
        """
        Load the CSV log file
        Args:
            logs_dir: Path to logs directory, defaults to self.config.logs_dir

        Returns:
            Logs dictionary
        """
        logs_dir = logs_dir or self.logs_dir
        csv_path = os.path.join(logs_dir, self.trainer_csv_log_file)
        logs = pd.read_csv(csv_path)
        return logs.to_dict()

    def prepare_input_batch(self, input_batch) -> Dict[str, torch.Tensor]:
        """
        Every operation required to prepare the inputs for model forward like moving to device, permutations, etc.

        Args:
            input_batch: Raw input batch from the dataloader

        Returns:
            The proper input batch required by model forward
        """
        # Put inputs on device manually if accelerator is not available, otherwise it's taken care of by the accelerator
        if self.accelerator is None:
            input_batch = {k: v.to(self.device) if isinstance(v, torch.Tensor) else v for k, v in input_batch.items()}

        return input_batch

    def amp_context_manager(self):
        """
        An auto context manager for mixed precision.

        Returns:
            A torch autocast context manager
        """
        if self.accelerator is not None:
            context_manager = self.accelerator.autocast()
        else:
            device_type = "cuda" if "cuda" in self.device else "cpu"
            dtype = torch.bfloat16 if self.config.mixed_precision == "bf16" or device_type == "cpu" else torch.float16
            enabled = self.config.mixed_precision is not None or self.config.mixed_precision == "no"
            context_manager = torch.autocast(
                device_type=device_type,
                dtype=dtype,
                enabled=enabled
            )
        return context_manager

    def forward(self, input_batch):
        """
        Perform model forward on the input batch

        In special cases, one can override this method in their desired trainer.

        Args:
            input_batch: Input batch

        Returns:
            Model outputs
        """
        if isinstance(input_batch, torch.Tensor):
            outputs = self.model(input_batch)
        elif isinstance(input_batch, Dict):
            forward_inputs = sanitize_function_parameters(self.model.forward, input_batch)
            outputs = self.model(**forward_inputs)
        else:
            raise ValueError(
                f"`input_batch` must be a tensor or a dict-like object containing key/value pairs of tensors, "
                f"but got {type(input_batch)}"
            )
        if not isinstance(outputs, Dict):
            raise ValueError(f"Model outputs must be dict-like not `{type(outputs)}`")

        return outputs

    def compute_loss(self, model_outputs: Dict, labels: torch.Tensor, **kwargs) -> torch.Tensor:
        """
        Compute loss from model outputs

        Args:
            model_outputs: Logits from model outputs
            labels: Ground truth labels

        Returns:
            The loss tensor
        """
        compute_loss_inputs = sanitize_function_parameters(self.model.compute_loss, model_outputs, **kwargs)
        compute_loss_inputs["labels"] = labels

        loss = self.model.compute_loss(**compute_loss_inputs)

        return loss

    def optimization_step(self):
        """
        Perform optimization step
        """
        if self.accelerator is not None:
            self.optimizer.step()
        else:
            self.scaler.step(self.optimizer)
            self.scaler.update()

        self.optimizer.zero_grad()

    def lr_scheduler_step(self, metrics=None):
        """
        Perform the learning rate scheduling step

        Args:
            metrics: one or multiple values that the scheduler watches to either perform step function or not. Only
             works for `ReduceLROnPlateau`.
        """
        if self.lr_scheduler is not None:
            if isinstance(self.lr_scheduler, lr_schedulers[LRSchedulerType.REDUCE_ON_PLATEAU]):
                self.lr_scheduler.step(metrics)
            else:
                self.lr_scheduler.step()

    def training_step(self, input_batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        """
        Train one batch of data and return loss and model outputs

        Args:
            input_batch: A batch of inputs to train

        Returns:
            Train step outputs including loss, logits, etc.
        """
        with self.amp_context_manager():
            outputs = self.forward(input_batch)

        loss = self.compute_loss(outputs, **input_batch) / self.config.gradient_accumulation_steps

        if self.accelerator is not None:
            self.accelerator.backward(loss)
        else:
            self.scaler.scale(loss).backward()

        outputs["loss"] = loss

        return outputs

    def evaluation_step(self, input_batch: Dict[str, torch.Tensor]) -> Dict[str, Any]:
        """
        Evaluate one batch of data and return loss and model outputs

        Args:
            input_batch: A batch of inputs to evaluate

        Returns:
            Evaluation step outputs including loss, logits, etc.
        """
        with self.amp_context_manager():
            outputs = self.forward(input_batch)

        loss = self.compute_loss(outputs, **input_batch)

        if self.model.is_generative and self.config.evaluate_with_generate:
            generate_inputs = sanitize_function_parameters(self.model.generate, input_batch)
            generated_ids = self.model.generate(**generate_inputs)
            outputs["logits"] = generated_ids["generated_ids"] if isinstance(generated_ids, dict) else generated_ids

        outputs["loss"] = loss

        return outputs

    def inner_training_loop(self, epoch_num: int):
        """
        Train the model for one epoch on the whole train dataset and verbose live metric values in the progress bar

        Args:
            epoch_num: Number of the current epoch

        Returns:
            Metrics averages through the full iteration
        """
        losses_sum = 0.0
        self.model.train()
        with tqdm(
            self.train_dataloader,
            unit="batch",
            desc=f"Epoch: {epoch_num}/{self.config.num_epochs} ",
            bar_format=TQDM_BAR_FORMAT,
            ascii=" #",
        ) as iterator:
            for step, input_batch in enumerate(iterator):
                input_batch = self.prepare_input_batch(input_batch)
                # Training on one batch
                outputs = self.training_step(input_batch)
                # Optimization step
                if (step + 1) % self.config.gradient_accumulation_steps == 0 or step == len(iterator):
                    self.optimization_step()
                # Gather outputs for metrics
                losses_sum += outputs["loss"].item()
                avg_loss = losses_sum / (step + 1)
                iterator.set_postfix(loss=avg_loss)
                self.state.global_step += 1

        return {"loss": avg_loss}

    def evaluate(self):
        """
        Evaluates the model on the whole eval dataset and verbose live metric values in the progress bar

        Returns:
            Evaluation results
        """
        self.metrics_handler.tracker.reset()
        self.model.eval()
        with tqdm(
            self.eval_dataloader,
            unit="batch",
            desc="Evaluating... ",
            bar_format=TQDM_BAR_FORMAT,
            ascii=" #",
        ) as iterator:
            with torch.inference_mode():
                for step, input_batch in enumerate(iterator):
                    input_batch = self.prepare_input_batch(input_batch)
                    # Evaluation on one batch
                    outputs = self.evaluation_step(input_batch)
                    logits = outputs["logits"].detach().cpu().numpy()
                    labels = input_batch["labels"].detach().cpu().numpy()
                    # Compute metrics
                    evaluation_results = self.metrics_handler.compute_metrics(logits, labels)
                    evaluation_results["loss"] = outputs["loss"].item()
                    # Gather outputs for metrics
                    self.metrics_handler.tracker.update(evaluation_results)
                    iterator.set_postfix(**self.metrics_handler.tracker.avg())

        return self.metrics_handler.tracker.avg()

    def print_info(self):
        """
        Print training info
        """

        def _print_info_line(key, value):
            line = f"  {colorize_text(key, 'bold')}: `{colorize_text(str(value), 'italic')}`"
            print(line)

        header = f"{'*' * 20} Training Info {'*' * 20}"
        footer = "*" * len(header)
        info = {
            "Output Directory": self.config.output_dir,
            "Task": self.config.task,
            "Model": type(self.model).__name__,
            "Init Weights": self.config.init_weights_from or "N/A",
            "Device(s)": self.device,
            "Batch Size": self.config.batch_size,
            "Epochs": self.config.num_epochs,
            "Training Dataset": self.train_dataset,
            "Evaluation Dataset": self.eval_dataset,
            "Optimizer": self.config.optimizer or self.default_optimizer,
            "Scheduler": self.config.lr_scheduler,
            "Initial Learning Rate": self.config.learning_rate,
            "Learning Rate Decay": self.config.weight_decay,
            "Number of Parameters": self.model.num_parameters,
            "Number of Trainable Parameters": self.model.num_trainable_parameters,
            "Mixed Precision": self.config.mixed_precision or "Full (fp32)",
            "Metrics": list(self.metrics_handler.metrics.keys()),
            "Checkpoints Path": self.checkpoints_dir,
            "Logs Path": self.logs_dir,
        }

        # Header
        print(f"\n{colorize_text(header, 'bold')}\n")
        # Info
        [_print_info_line(k, v) for k, v in info.items()]
        # Footer
        print(f"\n{colorize_text(footer, 'bold')}\n")

    def train(self, resume_from_checkpoint: str | bool = None):
        """
        The full training process like training, evaluation, logging and saving model checkpoints.

        Args:
            resume_from_checkpoint: Resume from checkpoint path (if value is a path) or automatically load from the
            latest checkpoint (if value is True)
        """
        if resume_from_checkpoint:
            self.load_from_checkpoint(resume_from_checkpoint)
            if self.current_epoch >= self.config.num_epochs:
                logger.info(
                    f"Unable to resume from `{os.path.join(self.checkpoints_dir, str(self.state.epoch))}` "
                    f"since it belongs to the ending epoch!"
                )
            self.current_epoch = self.state.epoch + 1

        self.print_info()

        for epoch in range(self.current_epoch, self.config.num_epochs + 1):
            print()

            # Train on the whole train data
            training_results = self.inner_training_loop(epoch)
            # Evaluate the model on eval data
            evaluation_results = self.evaluate()

            # LR scheduler step
            self.lr_scheduler_step(evaluation_results["loss"])

            # Metrics gathering
            train_logs = {f"train.{metric_name}": value for metric_name, value in training_results.items()}
            evaluation_logs = {f"evaluation.{metric_name}": value for metric_name, value in evaluation_results.items()}
            all_logs = {**train_logs, **evaluation_logs}

            # Update trainer state
            self.state.epoch = epoch
            self.state.update_best_results(
                metric_value=all_logs[self.config.metric_for_best_model],
                objective=self.metrics_handler.objective,
                step=epoch,
            )

            # maybe save checkpoint
            if epoch % self.config.save_freq == 0:
                ckpt_save_path = os.path.join(self.checkpoints_dir, str(epoch))
                self.save(ckpt_save_path)

            self.log(train_logs, evaluation_logs, epoch)

        logger.info("Training done!")

    def log(self, train_logs: Dict[str, Any], evaluation_logs: Dict[str, Any], step: int):
        """
        Log metrics results
        """
        # Log to tensorboard
        write_to_tensorboard(self.tensorboard, train_logs, step)
        write_to_tensorboard(self.tensorboard, evaluation_logs, step)

        # Log to CSV
        self.csv_logger.write({**train_logs, **evaluation_logs}, step)

        # Save trainer state
        self.state.save(
            os.path.join(
                self.checkpoints_dir,
                self.trainer_state_file,
            )
        )

    def save(
        self,
        path: str,
        config_filename=None,
        model_filename=None,
        model_config_filename=None,
        subfolder=None,
        dataset_config_file=None,
    ):
        """
        Save the trainer and relevant files to a path.

        Files to save are train config, model weights, model config, preprocessor files and preprocessor config.

        Args:
            path: A directory to save everything
            config_filename: Config filename
            model_filename: Model file name
            model_config_filename: Model config file name
            subfolder: Optional sub-folder
            dataset_config_file: Dataset config filename
        """
        config_filename = config_filename or self.trainer_config_file
        subfolder = subfolder or self.trainer_subfolder
        dataset_config_file = dataset_config_file or self.dataset_config_file

        self.config.save(path, filename=config_filename, subfolder=subfolder)
        self.model.save(path, filename=model_filename, config_filename=model_config_filename)
        if isinstance(self.train_dataset, Dataset):
            self.train_dataset.config.save(path, filename=dataset_config_file, subfolder=subfolder)
        else:
            logger.warning(
                f"The dataset passed to the Trainer is not a `hezar.data.Dataset` instance so that no dataset config"
                f" will be saved!"
            )

    def push_to_hub(
        self,
        repo_id: str,
        config_filename: str = None,
        push_model: bool = True,
        push_logs: bool = True,
        model_filename: str = None,
        model_config_filename: str = None,
        subfolder: str = None,
        dataset_config_filename: str = None,
        commit_message: str = None,
        private: bool = False,
    ):
        """
        Push everything to the Hub

        Args:
            repo_id: Path to hub
            config_filename: Trainer config file name
            push_model: Whether to push the model or not
            push_logs: Whether to push training logs or not
            model_filename: Model file name
            model_config_filename: Model config file name
            subfolder: Path to Trainer files
            dataset_config_filename: Dataset config file name
            commit_message: Commit message for the push
            private: Whether to create a private repo if it doesn't exist already
        """
        config_filename = config_filename or self.trainer_config_file
        subfolder = subfolder or self.trainer_subfolder
        dataset_config_file = dataset_config_filename or self.dataset_config_file

        # create remote repo
        create_repo(repo_id, repo_type="model", exist_ok=True, private=private)

        if not commit_message:
            commit_message = "Hezar: Upload training files"

        # upload train files
        self.config.push_to_hub(
            repo_id,
            filename=config_filename,
            subfolder=subfolder,
            private=private,
            commit_message=commit_message,
        )
        self.train_dataset.config.push_to_hub(
            repo_id,
            filename=dataset_config_file,
            subfolder=subfolder,
            private=private,
            commit_message=commit_message
        )

        # upload model
        if push_model:
            self.model.push_to_hub(
                repo_id,
                filename=model_filename,
                config_filename=model_config_filename,
                commit_message=commit_message,
                private=private,
            )

        if push_logs:
            upload_file(
                path_or_fileobj=self.csv_logger.save_path,
                path_in_repo=os.path.join(self.trainer_subfolder, self.trainer_csv_log_file),
                repo_id=repo_id,
                commit_message=commit_message,
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional

from omegaconf import DictConfig

from ..constants import PaddingType
from .logging import Logger


if TYPE_CHECKING:
    import torch


__all__ = [
    "convert_batch_dict_dtype",
    "resolve_inputs_length_for_padding",
    "pad_batch_items",
    "shift_tokens_right",
    "get_non_numeric_keys",
    "flatten_dict",
]

logger = Logger(__name__)


# TODO: This code might be able to be written in a cleaner way, but be careful, any change might break a lot of things!
def convert_batch_dict_dtype(batch_dict: Dict[str, Any], dtype: str = None, skip_keys: list = None):
    """
    Convert data dtypes of the values in a batch dict

    Args:
        batch_dict: The batched dictionary. Each key in the dict has a batch of data as its value.
        dtype: Target data type to convert to
        skip_keys: A list of key names to skip conversion

    Returns:
        The same dict with cast values
    """
    import numpy as np
    import torch

    dtype = dtype or "list"
    skip_keys = skip_keys or []

    if dtype == "list":
        for k, v in batch_dict.items():
            if isinstance(v, np.ndarray):
                batch_dict[k] = v.tolist()
            elif isinstance(v, torch.Tensor):
                batch_dict[k] = v.cpu().numpy().tolist()
        return batch_dict

    if dtype in ["np", "numpy"]:
        caster = np.asarray
        cast_type = np.ndarray
    elif dtype in ["pt", "torch", "pytorch"]:
        caster = torch.tensor
        cast_type = torch.Tensor
    else:
        raise ValueError(f"Invalid `dtype`: {dtype}")

    for k, v in batch_dict.items():
        if k not in skip_keys:
            try:
                if not isinstance(v, cast_type):
                    batch_dict[k] = caster(v)
            except Exception as e:  # noqa
                logger.warning(f"Could not convert values of `{k}` to type `{dtype}`\n" f"Error: {e}")
    return batch_dict


def resolve_inputs_length_for_padding(
    inputs: List[List[Any]],
    padding_type: str | PaddingType = None,
    max_length: Optional[bool | int] = None,
    truncation: Optional[bool] = True,
):
    """
    Resolve final inputs length based on padding_strategy and max_length values
    """
    inputs_max_length = max([len(x) for x in inputs])

    # Resolve padding and max_length values first
    if padding_type is None:
        if max_length is not None:
            padding_type = "max_length"
        else:
            padding_type = "longest"

    # Now lets resolve any conflicts
    if padding_type == "longest":
        if max_length is not None:
            logger.warning(
                "Setting padding='longest' and max_length is not valid. You must set one of them"
                " and leave the other as None. Falling back to padding='longest'"
            )

        inputs_length = inputs_max_length

    elif padding_type == "max_length":
        if max_length is None:
            logger.warning(
                "Setting padding='max_length' but no max_length value is provided! Falling back to padding='longest'"
            )
            inputs_length = inputs_max_length
        else:
            if max_length < inputs_max_length and not truncation:
                logger.warning(
                    f"Cannot set max_length to {max_length} "
                    f"while max input length is {inputs_max_length} and `truncation` is `False`"
                    f"Either set `truncation=True` or increase `max_length`"
                )
                inputs_length = inputs_max_length
            else:
                inputs_length = max_length
    else:
        raise ValueError(f"Invalid padding value `{padding_type}`, expected either `max_length` or `longest`")

    return inputs_length


def pad_batch_items(
    inputs: List[List[int | float]],
    padding_type: str | PaddingType = None,
    padding_side: Literal["right", "left"] = "right",
    pad_id: int = 0,
    max_length: Optional[bool | int] = None,
    truncation: Optional[bool] = True,
):
    """
    Given a nested container of unequal sized iterables e.g, batch of token ids, pad them based on padding strategy
    Args:
        inputs: A nested iterable of unequal sized iterables (e.g, list of lists)
        padding_type: Padding strategy, either max_length or longest
        padding_side: Where to add padding ids, `left` or `right`, defaults to `right`
        pad_id: Pad token id, defaults to `0`
        max_length: Max input length after padding, only applicable when padding_strategy == "max_length"
        truncation: Whether to truncate if an input in the batch is longer than max_length

    Returns:
        A list of equal sized lists
    """

    inputs_length = resolve_inputs_length_for_padding(
        inputs,
        padding_type=padding_type,
        max_length=max_length,
        truncation=truncation,
    )

    padded_inputs = []
    for ids in inputs:
        difference = inputs_length - len(ids)
        if difference > 0:
            paddings = [pad_id] * difference
            padded_ids = ids + paddings if padding_side == "right" else paddings + ids
            padded_inputs.append(padded_ids)
        else:
            padded_inputs.append(ids[:inputs_length])

    return padded_inputs


def shift_tokens_right(input_ids: "torch.Tensor", pad_token_id: int, decoder_start_token_id: int):
    """
    Shift input ids one token to the right.
    """
    shifted_input_ids = input_ids.new_zeros(input_ids.shape)
    shifted_input_ids[:, 1:] = input_ids[:, :-1].clone()
    shifted_input_ids[:, 0] = decoder_start_token_id

    # replace possible -100 values in labels by `pad_token_id`
    shifted_input_ids.masked_fill_(shifted_input_ids == -100, pad_token_id)

    return shifted_input_ids


def get_non_numeric_keys(d: Dict, batched=True):
    """
    Get keys that have string values in a dictionary

    Args:
        d: The dict
        batched: Are the input dict values batched or not

    Returns:
        A list of string-valued keys
    """
    keys = []
    for k, v in d.items():
        if len(v) and isinstance(v[0], list):
            if batched and not isinstance(v[0][0], (int, float, complex)) and not isinstance(v[0][0], bool):
                keys.append(k)
            elif isinstance(v[0], str):
                keys.append(k)
    return keys


def flatten_dict(dict_config: Dict | DictConfig) -> DictConfig:
    """
    Flatten a nested Dict/DictConfig object

    Args:
        dict_config: A Dict/DictConfig object

    Returns:
        The flattened version of the dict-like object
    """

    config = DictConfig({})
    for k, v in dict_config.items():
        if isinstance(v, (Dict, DictConfig)):
            config.update(flatten_dict(v))
        else:
            config[k] = v

    return config



from __future__ import annotations

from dataclasses import dataclass

from datasets import Audio, load_dataset

from .dataset import Dataset
from ..data_collators import SpeechRecognitionDataCollator
from ...configs import DatasetConfig
from ...constants import TaskType, Backends, PaddingType
from ...preprocessors import Tokenizer, AudioFeatureExtractor
from ...registry import register_dataset

_required_backends = [Backends.LIBROSA, Backends.DATASETS]


@dataclass
class SpeechRecognitionDatasetConfig(DatasetConfig):
    name = "speech_recognition"
    task = TaskType.SPEECH_RECOGNITION
    path: str = None
    feature_extractor_path: str = None
    tokenizer_path: str = None
    sampling_rate: int = 16000
    audio_array_padding_type: bool | str | PaddingType = None
    max_audio_array_length: int = None
    labels_padding_type: bool | str | PaddingType = None
    labels_max_length: int = None
    audio_file_path_column: str = "path"
    audio_column: str = "audio"
    audio_array_column: str = "array"
    transcript_column: str = "sentence"


@register_dataset("speech_recognition", config_class=SpeechRecognitionDatasetConfig)
class SpeechRecognitionDataset(Dataset):
    required_backends = _required_backends

    def __init__(self, config: SpeechRecognitionDatasetConfig, split=None, **kwargs):
        super().__init__(config, split, **kwargs)
        self.data = self._load(split)
        self.feature_extractor = AudioFeatureExtractor.load(self.config.feature_extractor_path)
        self.tokenizer = Tokenizer.load(self.config.tokenizer_path)
        self.data_collator = SpeechRecognitionDataCollator(
            self.feature_extractor,
            self.tokenizer,
            inputs_padding_type="max_length" if self.config.max_audio_array_length is not None else "longest",
            inputs_max_length=self.config.max_audio_array_length,
            labels_padding_type="max_length" if self.config.labels_max_length is not None else "longest",
            labels_max_length=self.config.labels_max_length,
        )

    def _load(self, split):
        data = load_dataset(self.config.path, split=split, cache_dir=self.cache_dir)
        data = data.cast_column(self.config.audio_column, Audio(sampling_rate=self.config.sampling_rate))
        return data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        sample_dict = self.data[index]
        transcript = sample_dict[self.config.transcript_column]
        audio_array = sample_dict[self.config.audio_column][self.config.audio_array_column]

        input_features = self.feature_extractor(
            audio_array,
            sampling_rate=self.config.sampling_rate,
            return_tensors="pt"
        )["input_features"]

        labels = self.tokenizer(
            transcript,
            max_length=self.config.labels_max_length,
            return_tensors="pt"
        )["token_ids"]

        return {
            "input_features": input_features,
            "labels": labels
        }



from __future__ import annotations

import os
from typing import List, Optional

from torch.utils.data import Dataset as TorchDataset

from ...builders import build_dataset
from ...configs import DatasetConfig
from ...constants import (
    DEFAULT_DATASET_CONFIG_FILE,
    HEZAR_CACHE_DIR,
    Backends,
    RepoType,
    SplitType,
)
from ...utils import verify_dependencies


class Dataset(TorchDataset):
    """
    Base class for all datasets in Hezar.

    Args:
        config: The configuration object for the dataset.
        **kwargs: Additional keyword arguments.

    Attributes:
        required_backends (List[str | Backends]): List of required backends for the dataset.
        config_filename (str): Default dataset config file name.
        cache_dir (str): Default cache directory for the dataset.

    """
    required_backends: List[str | Backends] = None
    config_filename = DEFAULT_DATASET_CONFIG_FILE
    cache_dir = os.path.join(HEZAR_CACHE_DIR, "datasets")

    def __init__(self, config: DatasetConfig, split=None, **kwargs):
        verify_dependencies(self, self.required_backends)
        self.config = config.update(kwargs)
        self.preprocessor = None
        self.data_collator = None
        self.split = split

    def __str__(self):
        dataset_name = self.config.path or self.config.name
        dataset_size = len(self)
        return f"{self.__class__.__name__}(path={dataset_name}['{self.split}'], size={dataset_size})"

    def __len__(self):
        """
        Returns the length of the dataset.

        Raises:
            NotImplementedError: This method must be implemented in derived classes.

        """
        raise NotImplementedError

    def __getitem__(self, index):
        """
        Gets a specific item from the dataset.

        Args:
            index: Index of the item to retrieve.

        Raises:
            NotImplementedError: This method must be implemented in derived classes.

        """
        raise NotImplementedError

    @classmethod
    def load(
        cls,
        hub_path: str | os.PathLike,
        config_filename: Optional[str] = None,
        split: Optional[str | SplitType] = None,
        cache_dir: str = None,
        **kwargs,
    ) -> "Dataset":
        """
        Load the dataset from a hub path.

        Args:
            hub_path (str | os.PathLike): Path to dataset from hub or locally.
            config_filename (Optional[str]): Dataset config file name.
            split (Optional[str | SplitType]): Dataset split, defaults to "train".
            cache_dir (str): Path to cache directory
            **kwargs: Config parameters as keyword arguments.

        Returns:
            Dataset: An instance of the loaded dataset.

        """
        split = split or "train"
        config_filename = config_filename or cls.config_filename
        if cache_dir is not None:
            cls.cache_dir = cache_dir
        dataset_config = DatasetConfig.load(
            hub_path,
            filename=config_filename,
            repo_type=RepoType.DATASET,
            cache_dir=cls.cache_dir,
        )
        dataset_config.path = hub_path
        dataset = build_dataset(dataset_config.name, config=dataset_config, split=split, **kwargs)
        return dataset
